# __version__.py
# autogenerated by keats 0.2.28
__version__ = "0.0.9"
__name__ = "dasi"
__title__ = "dasi"
__authors__ = ["Justin Vrana <justin.vrana@gmail.com>"]
__repo__ = None
__homepage__ = "https://github.com/jvrana/dasi-dna-design"
__description__ = "Automated DNA assembly planner for Python"
